@@ -2,32 +2,130 @@ use std::fs;
 use std::path;
 
 use lazy_static::lazy_static;
-use serde::{Deserialize, Serialize};
+use serde::{Deserialize, Deserializer, Serialize, Serializer};
+
+use crate::format::FlatpakManifestFormat;
+
+pub const ARCHIVE: &str = "archive";
+pub const GIT: &str = "git";
+pub const BAZAAR: &str = "bzr";
+pub const SVN: &str = "svn";
+pub const DIR: &str = "dir";
+pub const FILE: &str = "file";
+pub const SCRIPT: &str = "script";
+pub const SHELL: &str = "shell";
+pub const PATCH: &str = "patch";
+pub const EXTRA_DATA: &str = "extra-data";
 
 lazy_static! {
     pub static ref SOURCE_TYPES: Vec<String> = vec![
-        "archive".to_string(),
-        "git".to_string(),
-        "bzr".to_string(),
-        "svn".to_string(),
-        "dir".to_string(),
-        "file".to_string(),
-        "script".to_string(),
-        "shell".to_string(),
-        "patch".to_string(),
-        "extra-data".to_string(),
+        ARCHIVE.to_string(),
+        GIT.to_string(),
+        BAZAAR.to_string(),
+        SVN.to_string(),
+        DIR.to_string(),
+        FILE.to_string(),
+        SCRIPT.to_string(),
+        SHELL.to_string(),
+        PATCH.to_string(),
+        EXTRA_DATA.to_string(),
     ];
     pub static ref CODE_TYPES: Vec<String> = vec![
-        "archive".to_string(),
-        "git".to_string(),
-        "bzr".to_string(),
-        "svn".to_string(),
-        "dir".to_string(),
+        ARCHIVE.to_string(),
+        GIT.to_string(),
+        BAZAAR.to_string(),
+        SVN.to_string(),
+        DIR.to_string(),
     ];
-    pub static ref VCS_TYPES: Vec<String> = vec!["git".to_string(), "bzr".to_string(), "svn".to_string(),];
+    pub static ref VCS_TYPES: Vec<String> = vec![GIT.to_string(), BAZAAR.to_string(), SVN.to_string(),];
 }
 
-pub const DEFAULT_SOURCE_TYPE: &str = "archive";
+#[derive(Clone)]
+#[derive(Deserialize)]
+#[derive(Debug)]
+#[derive(Hash)]
+pub enum FlatpakSourceType {
+    Archive,
+    Git,
+    Bazaar,
+    Svn,
+    Dir,
+    File,
+    Script,
+    Shell,
+    Patch,
+    ExtraData,
+}
+impl Default for FlatpakSourceType {
+    fn default() -> Self {
+        FlatpakSourceType::Archive
+    }
+}
+impl FlatpakSourceType {
+    pub fn to_string(&self) -> String {
+        match &self {
+            FlatpakSourceType::Archive => ARCHIVE.to_string(),
+            FlatpakSourceType::Git => GIT.to_string(),
+            FlatpakSourceType::Bazaar => BAZAAR.to_string(),
+            FlatpakSourceType::Svn => SVN.to_string(),
+            FlatpakSourceType::Dir => DIR.to_string(),
+            FlatpakSourceType::File => FILE.to_string(),
+            FlatpakSourceType::Script => SCRIPT.to_string(),
+            FlatpakSourceType::Shell => SHELL.to_string(),
+            FlatpakSourceType::Patch => PATCH.to_string(),
+            FlatpakSourceType::ExtraData => EXTRA_DATA.to_string(),
+        }
+    }
+    pub fn from_string(source_type: &str) -> Result<FlatpakSourceType, String> {
+        if source_type == ARCHIVE {
+            return Ok(FlatpakSourceType::Archive);
+        }
+        if source_type == GIT {
+            return Ok(FlatpakSourceType::Git);
+        }
+        if source_type == BAZAAR {
+            return Ok(FlatpakSourceType::Bazaar);
+        }
+        if source_type == SVN {
+            return Ok(FlatpakSourceType::Svn);
+        }
+        if source_type == DIR {
+            return Ok(FlatpakSourceType::Dir);
+        }
+        if source_type == FILE {
+            return Ok(FlatpakSourceType::File);
+        }
+        if source_type == SCRIPT {
+            return Ok(FlatpakSourceType::Script);
+        }
+        if source_type == SHELL {
+            return Ok(FlatpakSourceType::Shell);
+        }
+        if source_type == PATCH {
+            return Ok(FlatpakSourceType::Patch);
+        }
+        if source_type == EXTRA_DATA {
+            return Ok(FlatpakSourceType::ExtraData);
+        }
+        Err(format!("Invalid source type {}.", source_type))
+    }
+}
+
+pub fn serialize_to_string<S>(x: &FlatpakSourceType, s: S) -> Result<S::Ok, S::Error>
+where
+    S: Serializer,
+{
+    s.serialize_str(&x.to_string())
+}
+
+pub fn deserialize_from_string<'de, D>(deserializer: D) -> Result<FlatpakSourceType, D::Error>
+where
+    D: Deserializer<'de>,
+{
+    let buf = String::deserialize(deserializer)?;
+
+    FlatpakSourceType::from_string(&buf).map_err(serde::de::Error::custom)
+}
 
 #[derive(Clone)]
 #[derive(Deserialize)]
@@ -43,114 +141,9 @@ pub const DEFAULT_SOURCE_TYPE: &str = "archive";
 /// Additionally, the sources list can contain a plain string, which is interpreted as the name
 /// of a separate json or yaml file that is read and inserted at this
 /// point. The file can contain a single source, or an array of sources.
-pub enum FlatpakSource {
+pub enum FlatpakSourceItem {
     Path(String),
-    Description(FlatpakSourceDescription),
-}
-impl FlatpakSource {
-    pub fn get_url(&self) -> Option<String> {
-        let source_description = match self {
-            FlatpakSource::Path(_) => return None,
-            FlatpakSource::Description(sd) => sd,
-        };
-        match &source_description.url {
-            Some(s) => Some(s.to_string()),
-            None => None,
-        }
-    }
-
-    pub fn get_all_mirror_urls(&self) -> Vec<String> {
-        let mut response: Vec<String> = vec![];
-
-        let source_description = match self {
-            FlatpakSource::Path(_) => return response,
-            FlatpakSource::Description(sd) => sd,
-        };
-        if let Some(urls) = &source_description.mirror_urls {
-            for url in urls {
-                response.push(url.to_string());
-            }
-        }
-        return response;
-    }
-
-    pub fn get_all_urls(&self) -> Vec<String> {
-        let mut response: Vec<String> = vec![];
-        let source_description = match self {
-            FlatpakSource::Path(_) => return response,
-            FlatpakSource::Description(sd) => sd,
-        };
-        if let Some(url) = &source_description.url {
-            response.push(url.to_string());
-        }
-        if let Some(urls) = &source_description.mirror_urls {
-            for url in urls {
-                response.push(url.to_string());
-            }
-        }
-        return response;
-    }
-
-    pub fn get_type_name(&self) -> String {
-        return match self {
-            FlatpakSource::Path(_) => "path".to_string(),
-            FlatpakSource::Description(d) => {
-                if let Some(t) = &d.r#type {
-                    return t.to_string();
-                }
-                return "empty".to_string();
-            }
-        };
-    }
-
-    pub fn has_commit(&self) -> bool {
-        return match self {
-            FlatpakSource::Path(_) => false,
-            FlatpakSource::Description(d) => d.commit.is_some(),
-        };
-    }
-
-    pub fn has_tag(&self) -> bool {
-        return match self {
-            FlatpakSource::Path(_) => false,
-            FlatpakSource::Description(d) => d.tag.is_some(),
-        };
-    }
-
-    pub fn has_branch(&self) -> bool {
-        return match self {
-            FlatpakSource::Path(_) => false,
-            FlatpakSource::Description(d) => d.branch.is_some(),
-        };
-    }
-
-    pub fn type_is_valid(&self) -> bool {
-        return match self {
-            FlatpakSource::Path(_) => true,
-            FlatpakSource::Description(d) => {
-                if let Some(t) = &d.r#type {
-                    return SOURCE_TYPES.contains(&t);
-                }
-                return false;
-            }
-        };
-    }
-
-    pub fn type_is_empty(&self) -> bool {
-        return match self {
-            FlatpakSource::Path(_) => false,
-            FlatpakSource::Description(d) => d.r#type.is_none(),
-        };
-    }
-
-    pub fn supports_mirror_urls(&self) -> bool {
-        let type_name = self.get_type_name();
-        // FIXME why are mirror urls not supported for types git, svn and bzr.
-        if type_name == "archive" || type_name == "file" {
-            return true;
-        }
-        return false;
-    }
+    Description(FlatpakSource),
 }
 
 #[derive(Clone)]
@@ -163,7 +156,7 @@ impl FlatpakSource {
 /// These contain a pointer to the source that will be extracted into the
 /// source directory before the build starts. They can be of several types,
 /// distinguished by the type property.
-pub struct FlatpakSourceDescription {
+pub struct FlatpakSource {
     /// Defines the type of the source description.
     /// It is not explicit in the flatpak-manifest man page,
     /// but we only found 1 source in all our dataset with an empty
@@ -346,46 +339,11 @@ pub struct FlatpakSourceDescription {
     #[serde(skip_serializing_if = "Option::is_none")]
     pub x_checker_data: Option<FlatpakDataCheckerConfig>,
 }
-impl FlatpakSourceDescription {
-    // The logic of this functions is based on the code from BuilderArchiveType.get_type
-    // in the flatpak-builder project.
-    pub fn detect_archive_type(url: &str) -> Option<String> {
-        let url = url.to_lowercase();
-        if url.ends_with(".tar") {
-            return Some("tar".to_string());
-        }
-        if url.ends_with(".tar.gz") || url.ends_with(".tgz") || url.ends_with(".taz") {
-            return Some("tar-gzip".to_string());
-        }
-        if url.ends_with(".tar.z") || url.ends_with(".taz") {
-            return Some("tar-compress".to_string());
-        }
-        if url.ends_with(".tar.bz2") || url.ends_with(".tz2") {
-            return Some("tar-bzip2".to_string());
-        }
-        if url.ends_with(".tbz2") || url.ends_with(".tbz") {
-            return Some("tar-bzip2".to_string());
-        }
-        if url.ends_with(".tar.lz") {
-            return Some("tar-lzip".to_string());
-        }
-        if url.ends_with(".tar.lzma") || url.ends_with(".tlz") {
-            return Some("tar-lzma".to_string());
-        }
-        if url.ends_with(".tar.lzo") {
-            return Some("tar-lzop".to_string());
-        }
-        if url.ends_with(".tar.xz") || url.ends_with(".txz") {
-            return Some("tar-xz".to_string());
-        }
-        if url.ends_with(".zip") {
-            return Some("zip".to_string());
-        }
-        if url.ends_with(".rpm") {
-            return Some("rpm".to_string());
-        }
-        if url.ends_with(".7z") {
-            return Some("sevenz".to_string());
+impl FlatpakSource {
+    /// Get the type for the Flatpak source.
+    pub fn get_type(&self) -> Option<String> {
+        if let Some(t) = &self.r#type {
+            return Some(t.clone());
         }
         None
     }
@@ -396,58 +354,58 @@ impl FlatpakSourceDescription {
         crate::filename::extension_is_valid(path)
     }
 
-    pub fn load_from_file(path: String) -> Result<Vec<FlatpakSourceDescription>, String> {
+    pub fn load_from_file(path: String) -> Result<Vec<FlatpakSource>, String> {
         let file_path = path::Path::new(&path);
         if !file_path.is_file() {
             return Err(format!("{} is not a file.", path));
         }
 
-        if FlatpakSourceDescription::file_path_matches(&file_path.to_str().unwrap()) {
-            let source_content = match fs::read_to_string(file_path) {
-                Ok(content) => content,
-                Err(e) => {
-                    return Err(format!(
-                        "Could not read file {}: {}!",
-                        file_path.to_str().unwrap(),
-                        e
-                    ))
-                }
-            };
+        let manifest_format = match FlatpakManifestFormat::from_path(&path) {
+            Some(f) => f,
+            None => return Err(format!("{} is not a Flatpak source manifest.", path)),
+        };
 
-            // A standalone source manifest can contain a single source, or an array
-            // of sources!!
-            if let Ok(source) = FlatpakSourceDescription::parse(&path, &source_content) {
-                return Ok(vec![source]);
-            }
-            if let Ok(sources) = FlatpakSourceDescription::parse_many(&path, &source_content) {
-                return Ok(sources);
+        let manifest_content = match fs::read_to_string(file_path) {
+            Ok(content) => content,
+            Err(e) => {
+                return Err(format!(
+                    "Could not read file {}: {}!",
+                    file_path.to_str().unwrap(),
+                    e
+                ))
             }
+        };
 
-            return Err(format!("Failed to parse Flatpak source at {}.", path));
-        } else {
-            return Err(format!("{} is not a Flatpak module.", path));
+        // A standalone source manifest can contain a single source, or an array
+        // of sources!!
+        if let Ok(source) = FlatpakSource::parse(manifest_format.clone(), &manifest_content) {
+            return Ok(vec![source]);
         }
+        if let Ok(sources) = FlatpakSource::parse_many(manifest_format, &manifest_content) {
+            return Ok(sources);
+        }
+
+        return Err(format!("Failed to parse Flatpak source manifest at {}.", path));
     }
 
-    pub fn parse(source_path: &str, source_content: &str) -> Result<FlatpakSourceDescription, String> {
-        let mut flatpak_source: FlatpakSourceDescription = FlatpakSourceDescription::default();
-
-        if source_path.to_lowercase().ends_with("yaml") || source_path.to_lowercase().ends_with("yml") {
-            flatpak_source = match serde_yaml::from_str(&source_content) {
+    pub fn parse(format: FlatpakManifestFormat, manifest_content: &str) -> Result<FlatpakSource, String> {
+        let mut flatpak_source: FlatpakSource = match &format {
+            FlatpakManifestFormat::YAML => match serde_yaml::from_str(&manifest_content) {
                 Ok(m) => m,
                 Err(e) => {
-                    return Err(format!("Failed to parse the Flatpak manifest: {}.", e));
+                    return Err(format!("Failed to parse the Flatpak source manifest: {}.", e));
                 }
-            };
-        } else if source_path.to_lowercase().ends_with("json") {
-            let json_content_without_comments = crate::utils::remove_comments_from_json(source_content);
-            flatpak_source = match serde_json::from_str(&json_content_without_comments) {
-                Ok(m) => m,
-                Err(e) => {
-                    return Err(format!("Failed to parse the Flatpak manifest: {}.", e));
+            },
+            FlatpakManifestFormat::JSON => {
+                let json_content_without_comments = crate::utils::remove_comments_from_json(manifest_content);
+                match serde_json::from_str(&json_content_without_comments) {
+                    Ok(m) => m,
+                    Err(e) => {
+                        return Err(format!("Failed to parse the Flatpak source manifest: {}.", e));
+                    }
                 }
-            };
-        }
+            }
+        };
 
         if let Err(e) = flatpak_source.is_valid() {
             return Err(e);
@@ -456,27 +414,26 @@ impl FlatpakSourceDescription {
     }
 
     pub fn parse_many(
-        source_path: &str,
-        source_content: &str,
-    ) -> Result<Vec<FlatpakSourceDescription>, String> {
-        let mut flatpak_sources: Vec<FlatpakSourceDescription> = vec![];
-
-        if source_path.to_lowercase().ends_with("yaml") || source_path.to_lowercase().ends_with("yml") {
-            flatpak_sources = match serde_yaml::from_str(&source_content) {
+        format: FlatpakManifestFormat,
+        manifest_content: &str,
+    ) -> Result<Vec<FlatpakSource>, String> {
+        let mut flatpak_sources: Vec<FlatpakSource> = match &format {
+            FlatpakManifestFormat::YAML => match serde_yaml::from_str(&manifest_content) {
                 Ok(m) => m,
                 Err(e) => {
-                    return Err(format!("Failed to parse the Flatpak manifest: {}.", e));
+                    return Err(format!("Failed to parse the Flatpak source manifest: {}.", e));
                 }
-            };
-        } else if source_path.to_lowercase().ends_with("json") {
-            let json_content_without_comments = crate::utils::remove_comments_from_json(source_content);
-            flatpak_sources = match serde_json::from_str(&json_content_without_comments) {
-                Ok(m) => m,
-                Err(e) => {
-                    return Err(format!("Failed to parse the Flatpak manifest: {}.", e));
+            },
+            FlatpakManifestFormat::JSON => {
+                let json_content_without_comments = crate::utils::remove_comments_from_json(manifest_content);
+                match serde_json::from_str(&json_content_without_comments) {
+                    Ok(m) => m,
+                    Err(e) => {
+                        return Err(format!("Failed to parse the Flatpak source manifest: {}.", e));
+                    }
                 }
-            };
-        }
+            }
+        };
 
         for flatpak_source in &flatpak_sources {
             if let Err(e) = flatpak_source.is_valid() {
@@ -487,9 +444,6 @@ impl FlatpakSourceDescription {
     }
 
     pub fn is_valid(&self) -> Result<(), String> {
-        if self.r#type.is_none() {
-            return Err("Required top-level field type is missing from Flatpak source.".to_string());
-        }
         if let Some(source_type) = &self.r#type {
             if !SOURCE_TYPES.contains(&source_type) {
                 return Err(format!("Invalid source type {}.", source_type));
@@ -497,6 +451,52 @@ impl FlatpakSourceDescription {
         }
         Ok(())
     }
+
+    pub fn get_url(&self) -> Option<String> {
+        match &self.url {
+            Some(s) => Some(s.to_string()),
+            None => None,
+        }
+    }
+
+    pub fn get_all_mirror_urls(&self) -> Vec<String> {
+        let mut response: Vec<String> = vec![];
+        if let Some(urls) = &self.mirror_urls {
+            for url in urls {
+                response.push(url.to_string());
+            }
+        }
+        return response;
+    }
+
+    pub fn get_all_urls(&self) -> Vec<String> {
+        let mut response: Vec<String> = vec![];
+        if let Some(url) = &self.url {
+            response.push(url.to_string());
+        }
+        if let Some(urls) = &self.mirror_urls {
+            for url in urls {
+                response.push(url.to_string());
+            }
+        }
+        return response;
+    }
+
+    pub fn get_type_name(&self) -> String {
+        if let Some(t) = &self.r#type {
+            return t.to_string();
+        }
+        return "empty".to_string();
+    }
+
+    pub fn supports_mirror_urls(&self) -> bool {
+        let type_name = self.get_type_name();
+        // FIXME why are mirror urls not supported for types git, svn and bzr.
+        if type_name == ARCHIVE || type_name == FILE {
+            return true;
+        }
+        return false;
+    }
 }
 
 #[derive(Clone)]
@@ -528,15 +528,15 @@ mod tests {
     use super::*;
 
     #[test]
-    pub fn test_parse_single_source() {
-        match FlatpakSourceDescription::parse(
-            "source.yaml",
+    pub fn test_parse_single_source_manifest() {
+        match FlatpakSource::parse(
+            FlatpakManifestFormat::YAML,
             r###"
             type: file
             path: apply_extra.sh
             "###,
         ) {
-            Err(e) => panic!(e),
+            Err(e) => std::panic::panic_any(e),
             Ok(source) => {
                 assert_eq!(source.path, Some("apply_extra.sh".to_string()));
             }
@@ -544,9 +544,9 @@ mod tests {
     }
 
     #[test]
-    pub fn test_parse_multiple_sources() {
-        match FlatpakSourceDescription::parse_many(
-            "source.yaml",
+    pub fn test_parse_multiple_source_manifests() {
+        match FlatpakSource::parse_many(
+            FlatpakManifestFormat::YAML,
             r###"
             - type: file
               path: apply_extra.sh
@@ -572,7 +572,7 @@ mod tests {
                 url-pattern: download_link_deb\s*=\s*"(http[s]?://[\w\d$-_@.&+]+)"
             "###,
         ) {
-            Err(e) => panic!(e),
+            Err(e) => std::panic::panic_any(e),
             Ok(sources) => {
                 assert_eq!(sources.len(), 4);
                 let last_source = sources.last().unwrap();
@@ -580,4 +580,39 @@ mod tests {
             }
         }
     }
+
+    #[test]
+    pub fn test_parse_invalid_source_type() {
+        let source_manifest = r###"
+            type: not_a_valid_source_type
+            path: apply_extra.sh
+        "###;
+        match FlatpakSource::parse(FlatpakManifestFormat::YAML, source_manifest) {
+            Ok(_source) => {
+                panic!("We should not be able to parse a source manifest with an invalid source type");
+            }
+            Err(e) => {
+                assert!(e.to_string().contains("Invalid source type"));
+            }
+        }
+    }
+
+    #[test]
+    pub fn test_parse_missing_source_type() {
+        let source_manifest = r###"
+            url: "https://ftp.gnu.org/gnu/gcc/gcc-7.5.0/gcc-7.5.0.tar.xz"
+            sha256: "b81946e7f01f90528a1f7352ab08cc602b9ccc05d4e44da4bd501c5a189ee661"
+        "###;
+        match FlatpakSource::parse(FlatpakManifestFormat::YAML, source_manifest) {
+            Ok(source) => {
+                assert!(source.url.is_some());
+            }
+            Err(e) => {
+                panic!(
+                    "We should be able to parse a source manifest without a source type: {}",
+                    e
+                );
+            }
+        }
+    }
 }
